import os
import shutil
import unittest
import textwrap

from sunset.api import _reset, BaseSettingsMissing, RolesNoMatch, hostname_like
from sunset.roles import BaseRole

STANDARD_SETTINGS = ['TEMPLATE_DIRS', 'USE_L10N', 'MANAGERS', 'MEDIA_ROOT',
                     'SITE_ID', 'DATABASES', 'MIDDLEWARE_CLASSES',
                     'ADMIN_MEDIA_PREFIX', 'TEMPLATE_LOADERS',
                     'ROOT_URLCONF', 'ADMINS', 'TEMPLATE_DEBUG',
                     'LANGUAGE_CODE', 'DEBUG', 'TIME_ZONE', 'SECRET_KEY',
                     'MEDIA_URL', 'USE_I18N', 'INSTALLED_APPS']


class ApiTest(unittest.TestCase):
    temporary_projects = []

    def setUp(self):
        _reset()
        self.project_path, self.settingsfunc = self.create_django_project()

    def tearDown(self):
        shutil.rmtree(self.project_path)

    def hostname(self):
        return os.uname()[1]

    def create_django_project(self):
        project_name = self.id().split('.')[-1]
        fixtures = os.path.join(os.path.dirname(__file__), 'fixtures')
        working = os.path.join(os.path.dirname(__file__), 'working')
        tpl_project = os.path.join(fixtures, 'tpl_project')
        project = os.path.join(working, project_name)

        if os.path.isdir(project):
            # Perhaps a bad test run left this
            shutil.rmtree(project)

        shutil.copytree(tpl_project, project)

        def get_settings():
            settings_module = __import__(
                'sunset.tests.working.%s.settings' % project_name,
                 globals(), locals(), ['*'])
            plain_settings = {}
            for setting in dir(settings_module):
                if setting == setting.upper():
                    plain_settings[setting] = getattr(settings_module, setting)
            return plain_settings

        return project, get_settings
        
    def make_settings(self, source):
        with open(os.path.join(self.project_path, 'settings.py'), 'w') as fh:
            fh.write(source)

    @property
    def current_settings(self):
        return self.settingsfunc()

    def assertStandardSettings(self, settings):
        for s in STANDARD_SETTINGS:
            self.assertTrue(s in settings)

    def test_can_read_base_settings(self):
        self.make_settings(textwrap.dedent("""
            from sunset import api

            import settingsbase
            api.collect(settingsbase)

            from sunset.collection import *
            """))

        settings = self.current_settings

        self.assertStandardSettings(settings)

    def test_can_set_dev_settings(self):
        self.make_settings(textwrap.dedent("""
            from sunset import api

            import settingsbase
            import settingsdev

            api.collect(settingsbase)
            api.dev_template(settingsdev)

            from sunset.collection import *
            """))

        settings = self.current_settings

        # The settingslocal file should not be created until roles is
        # called
        self.assertFalse(os.path.isfile(os.path.join(
            self.project_path, 'settingslocal.py')))
        self.assertStandardSettings(settings)

    def test_cannot_set_dev_without_base(self):
        self.make_settings(textwrap.dedent("""
            from sunset import api

            import settingsdev

            api.dev_template(settingsdev)

            from sunset.collection import *
            """))

        with self.assertRaises(BaseSettingsMissing) as ec:
            settings = self.current_settings

    def test_base_roll_can_match_hostname(self):
        role1 = BaseRole('non-matching-hostname')

        self.assertFalse(role1.hostname_matches('hostname1.local'))

        role2 = BaseRole('hostname')

        self.assertTrue(role2.hostname_matches('hostname1.local'))
        self.assertTrue(role2.hostname_matches('hostname2.local'))

    def test_api_hostname_like(self):
        hostname = self.hostname()

        if len(hostname) < 3:
            # This hostname is too short to run this test, skip it
            return

        self.assertTrue(hostname_like(self.hostname()))
        self.assertTrue(hostname_like(
            self.hostname(), 'not-match', 'not-match-too'))
        self.assertTrue(hostname_like(self.hostname().upper()))
        self.assertTrue(hostname_like(self.hostname()[:3]))
        self.assertFalse(hostname_like('something-that-will-never-match'))

    def test_can_create_dev_settings(self):
        self.make_settings(textwrap.dedent("""
            from sunset import api

            import settingsbase
            import settingsdev

            api.collect(settingsbase)
            api.dev_template(settingsdev)

            api.roles(
                api.dev('%s'))

            from sunset.collection import *
            """ % self.hostname()))

        settings = self.current_settings

        self.assertTrue(os.path.isfile(os.path.join(
            self.project_path, 'settingslocal.py')))
        self.assertStandardSettings(settings)

        self.assertTrue(settings['SETTINGS_LOCAL'])

    def test_can_create_deployment_settings(self):
        self.make_settings(textwrap.dedent("""
            from sunset import api

            import settingsbase
            import settingsdev

            from deployments import web01

            api.collect(settingsbase)
            api.dev_template(settingsdev)

            api.roles(
                api.deployment('%s', web01))

            from sunset.collection import *
            """ % self.hostname()))

        settings = self.current_settings

        # Should not have local settings
        self.assertFalse(os.path.isfile(os.path.join(
            self.project_path, 'settingslocal.py')))
        self.assertStandardSettings(settings)

        self.assertTrue(settings['SETTINGS_WEB_01'])

    def test_deployment_settings_missing_module(self):
        self.make_settings(textwrap.dedent("""
            from sunset import api

            import settingsbase
            import settingsdev

            api.collect(settingsbase)
            api.dev_template(settingsdev)

            api.roles(
                api.deployment('%s'))

            from sunset.collection import *
            """ % self.hostname()))

        with self.assertRaises(TypeError) as ec:
            settings = self.current_settings

    def test_can_add_settings_to_all(self):
        self.make_settings(textwrap.dedent("""
            from sunset import api

            import settingsbase
            import settingsdev

            from deployments import web01
            from deployments import facebook
                                           
            api.collect(settingsbase)
            api.collect(facebook)

            api.dev_template(settingsdev)

            api.roles(
                api.deployment('%s', web01))

            from sunset.collection import *
            """ % self.hostname()))

        settings = self.current_settings

        # None of the hostname matched, so we should not have this setting
        self.assertFalse('FACEBOOK_KEY' in settings)

    def test_roles_do_not_match(self):
        self.make_settings(textwrap.dedent("""
            from sunset import api

            import settingsbase
            import settingsdev

            api.collect(settingsbase)
            api.dev_template(settingsdev)

            roles = (
                api.dev('hostname1-no-match'),
                api.dev('hostname2-no-match'),
                api.dev('hostname3-no-match'),
                api.dev('hostname4-no-match'))

            api.roles(*roles)

            from sunset.collection import *
            """))

        with self.assertRaises(RolesNoMatch) as ec:
            settings = self.current_settings

    def test_will_create_empty_settingslocal(self):
        self.make_settings(textwrap.dedent("""
            from sunset import api

            import settingsbase

            api.collect(settingsbase)

            api.roles(
                api.dev('%s'))

            from sunset.collection import *
            """ % self.hostname()))

        settings = self.current_settings

        settingslocal_path = os.path.join(
            self.project_path, 'settingslocal.py')

        with open(settingslocal_path) as fh:
            source = fh.read()

        self.assertEquals(
            "# Override specific settings for this development machine here",
            source)



from os.path import isfile, dirname, join
import inspect

__all__ = ('DevRole', 'DeploymentRole')


class BaseRole(object):
    def __init__(self, hostname):
        self.hostname = hostname

        self.base_module = None
        self.dev_template_module = None

    def hostname_matches(self, full_hostname):
        if self.hostname in full_hostname:
            return True

        return False

    def get_module(self):
        raise NotImplemented()

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.hostname,)


class DevRole(BaseRole):
    local_name = 'settingslocal.py'
    local_module = 'settingslocal'

    empty_local_settings = \
        "# Override specific settings for this development machine here"

    def _create_local_settings(self, local_settings_path, template_module):
        if template_module:
            source = inspect.getsource(template_module)
        else:
            source = self.empty_local_settings
        
        with open(local_settings_path, 'w') as l:
            l.write(source)

    def get_module(self):
        local_settings_path = join(dirname(self.base_module.__file__), self.local_name)

        if not isfile(local_settings_path):
            self._create_local_settings(local_settings_path, self.dev_template_module)

        project = '.'.join(self.base_module.__name__.split('.')[0:-1])
        module = __import__('%s.%s' % (project, self.local_module),
            globals(), locals(), ['*'])

        return module


class DeploymentRole(BaseRole):
    def __init__(self, hostname, module):
        super(DeploymentRole, self).__init__(hostname)
        self.module = module

    def get_module(self):
        return self.module



from setuptools import setup, find_packages
import sys, os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
NEWS = open(os.path.join(here, 'NEWS.txt')).read()


version = '0.1'

install_requires = [
    # List your project dependencies here.
    # For more details, see:
    # http://packages.python.org/distribute/setuptools.html#declaring-dependencies
]


setup(name='django-sunset',
    version=version,
    description="Handle Django settings in a bit more organized fashion",
    long_description=README + '\n\n' + NEWS,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ],
    keywords='django settings',
    author='Rob Madole',
    author_email='robmadole@gmail.com',
    url='http://github.com/robmadole/django-sunset',
    license='BSD',
    packages=find_packages('src'),
    package_dir = {'': 'src'},include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    entry_points={}
)



import os

from sunset import collection
from sunset.roles import DevRole as dev
from sunset.roles import DeploymentRole as deployment

_base_module = None
_dev_template_module = None


def _reset():
    """
    Returns Django Sunset to an initial state, before anything has been done to
    it
    """
    global _base_module, _dev_template_module

    _base_module = None
    _dev_template_module = None

    for i in dir(collection):
        delattr(collection, i)
    

class BaseSettingsMissing(Exception):
    """
    For some operations, the base settings must be set first; raise this
    exception to indicate that the base settings are missing.
    """
    pass


class RolesNoMatch(Exception):
    """
    If no hostname could be found that matches the list of roles the user has
    defined, this should be raised
    """
    pass


def _extend_collection(module):
    """
    Takes the settings within the provided module and adds them to the
    collection that will eventually become the settings for this Django project
    """
    for setting in dir(module):
        if setting == setting.upper():
            setattr(collection, setting, getattr(module, setting))


def hostname_like(*hostnames):
    """
    Takes the cmp_hostname and determines if it matches the real hostname of
    this machine
    """
    for cmp_hostname in hostnames:
        real_hostname = os.uname()[1]
        if cmp_hostname.lower() in real_hostname.lower():
            return True
        return False


def collect(module):
    """
    Takes all the settings within the module and adds them to what will become
    the final settings for the Django project
    """
    global _base_module

    if not _base_module and 'SECRET_KEY' in dir(module):
        # Let's assume this is the base_module
        _base_module = module

    _extend_collection(module)


def dev_template(module):
    """
    Sunset supports the creation of a settingslocal.py file that is specific to
    the developer (that doesn't get checked into the code repository) and
    allows settings to be overridden.  The module argument here represents that
    settings template.
    """
    global _base_module, _dev_template_module

    if not _base_module:
        raise BaseSettingsMissing('Cannot set the development template '
            'module, the base has not been set yet')

    _dev_template_module = module


def roles(*roles):
    """
    Takes a list of roles and populates the collection according to which roles
    match the hostname.
    """
    global _base_module, _dev_template_module

    hostname = os.uname()[1]

    for node in roles:
        if node.hostname_matches(hostname):
            node.base_module = _base_module
            node.dev_template_module = _dev_template_module

            module = node.get_module()
            _extend_collection(module)

            return node

    # Getting to here means we did not match any hosts
    raise RolesNoMatch('We could not find any matches for hostname %s, you '
        'probably need to add this host into your settings' % hostname)
